@@ -1,8 +1,8 @@
 /* gxmessage - an xmessage clone using GTK2
  *
- * $Id: gxmessage.c,v 1.18 2005/02/22 09:22:38 trm Exp trm $
+ * $Id: gxmessage.c,v 1.23 2006/01/23 03:40:07 trm Exp trm $
  *
- * Copyright Timothy Musson 2001 - 2005, <trmusson@ihug.co.nz>
+ * Copyright Timothy Musson 2001 - 2006, <trmusson@ihug.co.nz>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,14 +18,13 @@
  *
  */
 
-
-
 #include <config.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
 #include <gtk/gtk.h>
+#include <gdk/gdkkeysyms.h>
 
 
 #ifdef ENABLE_NLS
@@ -51,14 +50,13 @@
                       "pixmaps" G_DIR_SEPARATOR_S "gxmessage.png"
 
 
-typedef struct _ButtonList ButtonList;
-
+typedef struct _Button Button;
 
 
 struct {
 	gchar             *message_text;
-	gint              message_length;
-	ButtonList        *button_list;
+	gint              message_len;
+	Button            *button_list;
 	const gchar       *default_str;
 	const gchar       *title_str;
 	const gchar       *geom_str;
@@ -81,12 +79,12 @@ struct {
 } gx;
 
 
-struct _ButtonList {
-	gboolean          is_default;
-	gint              value;
-	const gchar       *label_str;
-	ButtonList        *prev;
-	ButtonList        *next;
+struct _Button {
+	gboolean      is_default;
+	gint          value;
+	const gchar   *label;
+	Button        *prev;
+	Button        *next;
 };
 
 
@@ -97,7 +95,7 @@ struct Option {
 };
 
 
-static struct Option option[] = { /* min_len, requires_arg, opt_str */
+static struct Option option[] = {
 	{2, TRUE,  "buttons"},
 	{1, FALSE, "center"},
 	{2, TRUE,  "default"},
@@ -120,6 +118,8 @@ static struct Option option[] = { /* min_len, requires_arg, opt_str */
 	{2, TRUE,  "selectionTimeout"},
 	{2, FALSE, "synchronous"},
 	{2, TRUE,  "xnllanguage"},
+	{2, TRUE,  "name"},
+	{2, TRUE,  "display"},
 	{2, FALSE, "borderless"},
 	{1, FALSE, "wrap"},
 	{3, TRUE,  "encoding"},
@@ -156,6 +156,8 @@ enum {
 	OPT_SELECTIONTIMEOUT,
 	OPT_SYNCHRONOUS,
 	OPT_XNLLANGUAGE,
+	OPT_NAME,
+	OPT_DISPLAY,
 	OPT_BORDERLESS,
 	OPT_WRAP,
 	OPT_ENCODING,
@@ -169,13 +171,11 @@ enum {
 };
 
 
-
 void prog_cleanup (void);
 
 
-
-ButtonList*
-button_first (ButtonList *button)
+Button*
+button_first (Button *button)
 {
 	if (button != NULL) {
 		while (button->prev != NULL) {
@@ -186,11 +186,10 @@ button_first (ButtonList *button)
 }
 
 
-
 void
-button_free (ButtonList *button)
+button_free_all (Button *button)
 {
-	ButtonList *next;
+	Button *next;
 
 	button = button_first (button);
 	while (button != NULL) {
@@ -201,95 +200,88 @@ button_free (ButtonList *button)
 }
 
 
-
-ButtonList*
-button_append (ButtonList *button, ButtonList *new_button)
+Button*
+button_append (Button *button, Button *button_new)
 {
 	if (button != NULL) {
-		new_button->prev = button;
-		button->next = new_button;
+		button_new->prev = button;
+		button->next = button_new;
 	}
-	return new_button;
+	return button_new;
 }
 
 
-
-ButtonList*
-button_new_from_str (gchar *str, gint default_value, gint len)
+gint
+parse_label_value_pair (gchar *str, gint value, gint *len, gboolean *end)
 {
-	/* Given a string containing "label" or "label:value", fill in a
-	 * new button structure and return it.
-	 */
+	gchar *colon = NULL;
 
-	ButtonList *button = g_new0 (ButtonList, 1);
-	gboolean value_from_str = FALSE;
-	gint i = len;
+	*end = FALSE;
+	*len = 0;
 
+	while (*str != '\0') {
+		if (*str == '\\') {
+			/* unescape */
+			memmove (str, str + 1, strlen (str));
+		}
+		else if (*str == ':') {
+			/* take note of the last colon found */
+			colon = str;
+		}
+		else if (*str == ',') {
+			/* end of pair */
+			break;
+		}
+		str++;
+		*len = *len + 1;
+	}
 
-	/* parse "label:value" format in str */
-	str[len] = '\0';
-	while (i != 0 && !value_from_str) value_from_str = str[--i] == ':';
-	if (value_from_str) {
-		/* get label and value from str */
-		str[i] = '\0';
-		button->label_str = str;
-		button->value = atoi (&str[i + 1]);
+	if (*str == '\0') {
+		*end = TRUE;
 	}
 	else {
-		/* get label from str, value from default_value */
-		button->label_str = str;
-		button->value = default_value;
+		*str = '\0';
 	}
 
-	/* xmessage will only draw the buttons if at least one of
-	 * them has a label. To emulate that, keep track of it here:
-	 */
-	gx.do_buttons = gx.do_buttons || len > 0;
+	if (colon) {
+		/* replace default value with value from string */
+		*colon = '\0';
+		value = atoi (++colon);
+	}
 
-	return button;
+	return value;
 }
 
 
-
-ButtonList*
+Button*
 button_list_from_str (gchar *str)
 {
-	/* Given a string containing a comma-separated list of "label" and/or
-	 * "label:value", build and return a list of buttons. The return points
-	 * to the last button in the list.
-	 */
-
-	ButtonList *button = NULL;
-	ButtonList *b;
-	gint len;
-	gint i = 0;
-	gint startpos = 0;
-	gint value = 101;
+	/* Split "LABEL:VALUE,LABEL:VALUE,..." into a list of buttons */
 
+	Button *button, *blist = NULL;
+	gint len, value, default_value = 101;
+	gboolean end;
 
 	if (str == NULL) return NULL;
 
-	len = strlen (str);
-	while (i < len) {
-		if (str[i] == ',') {
-			b = button_new_from_str (&str[startpos], value, i - startpos);
-			button = button_append (button, b);
-			startpos = i + 1;
-			value++;
-		}
-		i++;
-	}
+	do {
+		value = parse_label_value_pair (str, default_value, &len, &end);
+		gx.do_buttons |= len > 0;
+		button = g_new0 (Button, 1);
+		button->label = str;
+		button->value = value;
+		blist = button_append (blist, button);
+		str = str + len + 1;
+		default_value++;
+	} while (!end);
 
-	b = button_new_from_str (&str[startpos], value, i - startpos);
-	button = button_append (button, b);
-
-	return button;
+	/* return the last item */
+	return blist;
 }
 
 
-
 void
-button_set_default (ButtonList *button, const gchar *str)
+button_set_default (Button *button, const gchar *str)
 {
 	/* Make button->is_default TRUE for each button whose label matches str */
 
@@ -297,13 +289,12 @@ button_set_default (ButtonList *button, const gchar *str)
 
 	button = button_first (button);
 	while (button != NULL) {
-		button->is_default = strcmp (button->label_str, str) == 0;
+		button->is_default = strcmp (button->label, str) == 0;
 		button = button->next;
 	}
 }
 
 
-
 gint
 my_get_opt (const gchar *str, gboolean not_last)
 {
@@ -313,30 +304,24 @@ my_get_opt (const gchar *str, gboolean not_last)
 	 * on the command line.
 	 */
 
-	gint opt;
-	gint len;
-	gint offset = 1;
-
+	gint opt, len;
 
 	if (strcmp (str, "+rv") == 0) return OPT_RV;
-	if (str[0] != '-') return OPT_IS_UNKNOWN;
+	if (*str != '-') return OPT_IS_UNKNOWN;
+
+	str++;
+	if (*str == '-') str++;
+	len = strlen (str);
 
-	len = strlen (&str[offset]);
 	if (len > 0) {
-		if (str[1] == '-') {
-			offset++;
-			len--;
-		}
-		if (len > 0) {
-			for (opt = 0; opt < N_OPTS; opt++) {
-				if (len >= option[opt].min_len &&
-				    strncmp (&str[offset], option[opt].opt_str, len) == 0) {
-					if (!option[opt].requires_arg || not_last) {
-						return opt;
-					}
-					else {
-						return OPT_IS_MISSING_ARG;
-					}
+		for (opt = 0; opt < N_OPTS; opt++) {
+			if (len >= option[opt].min_len &&
+			    strncmp (str, option[opt].opt_str, len) == 0) {
+				if (!option[opt].requires_arg || not_last) {
+					return opt;
+				}
+				else {
+					return OPT_IS_MISSING_ARG;
 				}
 			}
 		}
@@ -345,7 +330,6 @@ my_get_opt (const gchar *str, gboolean not_last)
 }
 
 
-
 void
 cb_window_destroy (GtkWidget *widget, GdkEvent *event, gpointer data)
 {
@@ -353,14 +337,23 @@ cb_window_destroy (GtkWidget *widget, GdkEvent *event, gpointer data)
 }
 
 
+gboolean
+cb_key_press (GtkWidget *w, GdkEventKey *event, gpointer data)
+{
+	if (event->keyval == GDK_Escape) {
+		gtk_main_quit ();
+	}
+	return FALSE;
+}
+
 
 void
 cb_button_clicked (GtkWidget *widget, gpointer data)
 {
-	gx.exit_code = ((ButtonList*)data)->value;
+	gx.exit_code = ((Button*)data)->value;
 
 	if (gx.do_print) {
-		g_print ("%s\n", ((ButtonList*)data)->label_str);
+		g_print ("%s\n", ((Button*)data)->label);
 	}
 	else if (gx.entry_str != NULL) {
 		g_print ("%s\n", gtk_entry_get_text (GTK_ENTRY(gx.entry_widget)));
@@ -369,7 +362,6 @@ cb_button_clicked (GtkWidget *widget, gpointer data)
 }
 
 
-
 void
 cb_entry_activated (GtkWidget *widget, gpointer data)
 {
@@ -379,8 +371,7 @@ cb_entry_activated (GtkWidget *widget, gpointer data)
 }
 
 
-
-gint
+gboolean
 cb_timeout (gpointer data)
 {
 	static gint counter = 0;
@@ -393,14 +384,14 @@ cb_timeout (gpointer data)
 }
 
 
-
 gint
 geometry_get_number (const gchar **geometry)
 {
+	/* (This function is taken from GNOME's gnome-geometry.c) */
+
 	gint value = 0;
 	gint mult  = 1;
 
-
 	if (**geometry == '-') {
 		mult = -1;
 		(*geometry)++;
@@ -413,116 +404,53 @@ geometry_get_number (const gchar **geometry)
 }
 
 
-
-gboolean
-geometry_parse_string (const gchar *geometry,
-                       gint *x, gint *y, gint *w, gint *h)
+void
+geometry_parse_string (const gchar *geometry, gint *width, gint *height)
 {
-	/* Parse geometry string into x, y, w and h. Return TRUE if successful,
-	 * else FALSE. (This function and geometry_get_number are taken from
-	 * the GNOME 1.x gnome-geometry.c)
-	 */
+	/* Parse geometry string into width and height */
 
-	gint subtract;
-
-
-	g_return_val_if_fail (x != NULL, FALSE);
-	g_return_val_if_fail (y != NULL, FALSE);
-	g_return_val_if_fail (w != NULL, FALSE);
-	g_return_val_if_fail (h != NULL, FALSE);
-
-	*x = *y = *w = *h = -1;
-
-	if (geometry == NULL) return FALSE;
+	*width = *height = -1;
 
 	if (*geometry == '=') geometry++;
-	if (!*geometry) return FALSE;
+	if (*geometry == '\0') return;
+
+	*width = geometry_get_number (&geometry);
+	if (*width < -1) *width = -1;
 
-	if (isdigit (*geometry)) {
-		*w = geometry_get_number (&geometry);
-	}
-	if (!*geometry) {
-		return TRUE;
-	}
 	if (*geometry == 'x') {
 		geometry++;
-		*h = geometry_get_number (&geometry);
+		if (*geometry == '\0') return;
+		*height = geometry_get_number (&geometry);
+		if (*height < -1) *height = -1;
 	}
-	if (!*geometry) {
-		return TRUE;
-	}
-
-
-	if (*geometry == '+') {
-		subtract = 0;
-		geometry++;
-	}
-	else if (*geometry == '-') {
-		subtract = gdk_screen_width ();
-		geometry++;
-	}
-	else {
-		return FALSE;
-	}
-
-
-	*x = geometry_get_number (&geometry);
-	if (subtract) {
-		*x = subtract - *x;
-	}
-	if (!*geometry) {
-		return TRUE;
-	}
-
-
-	if (*geometry == '+') {
-		subtract = 0;
-		geometry++;
-	}
-	else if (*geometry == '-') {
-		subtract = gdk_screen_height ();
-		geometry++;
-	}
-	else {
-		return FALSE;
-	}
-
-
-	*y = geometry_get_number (&geometry);
-	if (subtract) {
-		*y = subtract - *y;
-	}
-
-	return TRUE;
 }
 
 
-
 void
 window_create (void)
 {
 	GtkWidget  *window;
 	GtkWidget  *vbox, *vbox2;
 	GtkWidget  *scroller;
-	GtkWidget  *button_box = NULL;
-	GtkWidget  *widget = NULL;
+	GtkWidget  *btn_box;
+	GtkWidget  *btn;
 	GtkWidget  *message_widget;
-	ButtonList *button_list;
+	Button     *button;
 	GdkColor   color;
-	PangoFontDescription *font_desc;
 	GtkRequisition size_req;
 	GtkTextBuffer *buf;
-	gint win_w, win_h;
+	GtkTextIter iter;
+	gint       win_w, win_h;
+	gint       max_w, max_h;
 
 
 	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 
+	g_signal_connect (G_OBJECT(window), "destroy",
+	                  G_CALLBACK(cb_window_destroy), NULL);
 
-	g_signal_connect (G_OBJECT(window),
-	                  "destroy",
-	                  G_CALLBACK(cb_window_destroy),
-	                  NULL);
-
+	g_signal_connect (G_OBJECT(window), "key_press_event",
+	                  G_CALLBACK(cb_key_press), NULL);
 
 	if (gx.title_str != NULL) {
 		gtk_window_set_title (GTK_WINDOW(window), gx.title_str);
@@ -537,12 +465,8 @@ window_create (void)
 	if (gx.do_borderless) {
 		gtk_window_set_decorated (GTK_WINDOW(window), FALSE);
 	}
-#if _GTK_2_0
-#else
-	/* requires GTK >= 2.4.0 */
-	gtk_window_set_accept_focus (GTK_WINDOW(window), gx.do_focus);
-#endif
 
+	gtk_window_set_accept_focus (GTK_WINDOW(window), gx.do_focus);
 
 	/* window contents */
 	gtk_container_set_border_width (GTK_CONTAINER(window), 12);
@@ -561,27 +485,27 @@ window_create (void)
 	gtk_scrolled_window_set_placement (GTK_SCROLLED_WINDOW(scroller), GTK_CORNER_TOP_LEFT);
 	gtk_box_pack_start (GTK_BOX(vbox2), scroller, TRUE, TRUE, 0);
 
-
 	/* the message */
 	message_widget = gtk_text_view_new ();
+	gtk_widget_set_name (message_widget, "gxmessage-textview");
 	gtk_text_view_set_editable (GTK_TEXT_VIEW(message_widget), FALSE);
 	gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW(message_widget), TRUE);
 	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW(message_widget), gx.wrap_mode);
 
 	if (gx.font_str != NULL) {
+		PangoFontDescription *font_desc;
 		font_desc = pango_font_description_from_string (gx.font_str);
+		gtk_widget_modify_font (message_widget, font_desc);
+		pango_font_description_free (font_desc);
 	}
-	else {
-		/* default to a monospaced font */
-		font_desc = pango_font_description_from_string ("monospace");
-	}
-	gtk_widget_modify_font (message_widget, font_desc);
-	pango_font_description_free (font_desc);
 
 	gtk_container_add (GTK_CONTAINER(scroller), message_widget);
 
 	buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW(message_widget));
-	gtk_text_buffer_set_text (buf, gx.message_text, gx.message_length);
+	gtk_text_buffer_set_text (buf, gx.message_text, strlen (gx.message_text));
+
+	gtk_text_buffer_get_start_iter (buf, &iter);
+	gtk_text_buffer_place_cursor (buf, &iter);
 
 	if (gx.color_fg != NULL) {
 		if (gdk_color_parse (gx.color_fg, &color)) {
@@ -598,93 +522,100 @@ window_create (void)
 	/* text entry */
 	if (gx.entry_str != NULL) {
 		gx.entry_widget = gtk_entry_new ();
+		gtk_widget_set_name (gx.entry_widget, "gxmessage-entry");
 		gtk_editable_set_editable (GTK_EDITABLE(gx.entry_widget), TRUE);
 		gtk_entry_set_text (GTK_ENTRY(gx.entry_widget), gx.entry_str);
 		gtk_box_pack_start (GTK_BOX(vbox), gx.entry_widget, FALSE, FALSE, 5);
 		gtk_widget_grab_focus (gx.entry_widget);
 		if (!gx.do_buttons) {
 			/* allow hitting <RETURN> to close the window */
-			g_signal_connect (G_OBJECT(gx.entry_widget),
-			                  "activate",
-			                  G_CALLBACK(cb_entry_activated),
-			                  (gpointer)0);
+			g_signal_connect (G_OBJECT(gx.entry_widget), "activate",
+			                  G_CALLBACK(cb_entry_activated), (gpointer)0);
 		}
 	}
 
-
 	/* add buttons */
 	if (gx.do_buttons) {
 
-		button_list = button_first (gx.button_list);
+		button = button_first (gx.button_list);
 
-		button_box = gtk_hbutton_box_new ();
-		gtk_button_box_set_layout (GTK_BUTTON_BOX(button_box),
+		btn_box = gtk_hbutton_box_new ();
+		gtk_button_box_set_layout (GTK_BUTTON_BOX(btn_box),
 		                           GTK_BUTTONBOX_END);
-		gtk_box_set_spacing (GTK_BOX(button_box), 6);
-		gtk_box_pack_end (GTK_BOX(vbox), button_box, FALSE, FALSE, 0);
+		gtk_box_set_spacing (GTK_BOX(btn_box), 6);
+		gtk_box_pack_end (GTK_BOX(vbox), btn_box, FALSE, FALSE, 0);
 
-		while (button_list != NULL) {
+		while (button != NULL) {
 
-			if (g_str_has_prefix (button_list->label_str, "GTK_STOCK_")) {
+			if (g_str_has_prefix (button->label, "GTK_STOCK_")) {
 				gchar *s;
 				gchar *p;
-				p = g_ascii_strdown (button_list->label_str + 10, -1);
+				p = g_ascii_strdown (button->label + 10, -1);
 				s = p;
 				while (*s != '\0') {
 					if (*s == '_') *s = '-';
 					s++;
 				}
 				s = g_strconcat ("gtk-", p, NULL);
-				widget = gtk_button_new_from_stock (s);
+				btn = gtk_button_new_from_stock (s);
 				g_free (s);
 				g_free (p);
-				if (widget == NULL) {
-					widget = gtk_button_new_with_label (button_list->label_str);
-				}
 			}
 			else {
-				widget = gtk_button_new_with_mnemonic (button_list->label_str);
+				btn = gtk_button_new_with_mnemonic (button->label);
 			}
 
-			g_signal_connect (G_OBJECT(widget),
-			                  "clicked",
-			                  G_CALLBACK(cb_button_clicked),
-			                  (gpointer)button_list);
-			gtk_box_pack_start (GTK_BOX(button_box), widget, FALSE, FALSE, 0);
+			g_signal_connect (G_OBJECT(btn), "clicked",
+			                  G_CALLBACK(cb_button_clicked), (gpointer)button);
+			gtk_box_pack_start (GTK_BOX(btn_box), btn, FALSE, FALSE, 0);
 
-			if (button_list->is_default) {
-				gtk_widget_grab_focus (widget);
+			if (button->is_default) {
+				gtk_widget_grab_focus (btn);
 			}
 
-			button_list = button_list->next;
+			button = button->next;
 		}
-
 	}
 
-
 	/* window geometry */
 
-	win_w = 330;  /* XXX */
-	win_h = 90;   /* XXX */
+	max_w = gdk_screen_width () * 0.7;
+	max_h = gdk_screen_height () * 0.7;
+
+	if (gx.message_len > 3000) {
+		win_w = max_w;
+		win_h = max_h;
+	}
+	else {
+		/* render dummy text, to get an idea of its size */
+		GtkWidget *dummy = gtk_label_new (gx.message_text);
+		gtk_widget_modify_font (dummy,
+		    gtk_widget_get_style (message_widget)->font_desc);
+		gtk_container_add (GTK_CONTAINER(vbox), dummy);
+		gtk_widget_size_request (dummy, &size_req);
+		gtk_widget_destroy (dummy);
+		/* ~50 pixels for borders and scrollbar space */
+		win_w = size_req.width + 50;
+		win_h = size_req.height + 50;
+		if (win_w > max_w) win_w = max_w;
+		if (win_h > max_h) win_h = max_h;
+	}
 
 	if (gx.entry_str != NULL) {
-		gtk_widget_realize (gx.entry_widget);
 		gtk_widget_size_request (gx.entry_widget, &size_req);
-		win_h = win_h + size_req.height + 10;
+		win_h = win_h + size_req.height + 12;
 	}
+
 	if (gx.do_buttons) {
-		gtk_widget_size_request (button_box, &size_req);
-		win_h = win_h + size_req.height;
+		gtk_widget_size_request (btn, &size_req);
+		win_h = win_h + size_req.height + 12;
 	}
 
 	if (gx.geom_str != NULL) {
-		gint tmp_x, tmp_y, tmp_w, tmp_h;
-		if (geometry_parse_string (gx.geom_str, &tmp_x, &tmp_y, &tmp_w, &tmp_h)) {
-			if (win_w != -1) {
-				win_w = tmp_w;
-				win_h = tmp_h;
-			}
-		}
+		gint tmp_w, tmp_h;
+		geometry_parse_string (gx.geom_str, &tmp_w, &tmp_h);
+		if (tmp_w != -1) win_w = tmp_w;
+		if (tmp_h != -1) win_h = tmp_h;
 	}
 
 	gtk_window_set_default_size (GTK_WINDOW(window), win_w, win_h);
@@ -695,13 +626,12 @@ window_create (void)
 
 	/* begin timeout */
 	if (gx.timeout != 0) {
-		gx.timeout_id = gtk_timeout_add (1000, (GtkFunction)cb_timeout,
-		                                 (gpointer)gx.timeout);
+		gx.timeout_id = g_timeout_add (1000, (GtkFunction)cb_timeout,
+		                               (gpointer)gx.timeout);
 	}
 }
 
 
-
 gchar*
 read_stdin (void)
 {
@@ -720,7 +650,6 @@ read_stdin (void)
 }
 
 
-
 gchar*
 message_to_utf8 (const gchar *str)
 {
@@ -752,42 +681,43 @@ message_to_utf8 (const gchar *str)
 }
 
 
-
-void
-option_kludge (gint *argc, gchar ***argv)
+gboolean
+my_gtk_init (gint argc, gchar *argv[])
 {
-	/* Vet argv before gtk_init sees it */
+	/* Let gtk_init see --display and --name, but no other options.
+	 * Return FALSE if gtk_init fails.
+	 */
 
-	gint i, len;
+	gboolean ok;
+	gchar *s, **av;
+	gint i, len, n = 1;
 
-	for (i = 1; i < *argc; i++) {
-		len = strlen ((*argv)[i]);
-		if (len > 2) {
-			if (strncmp ("-display", (*argv)[i], len) == 0) {
-				(*argv)[i] = "--display";
-				if (i + 1 < *argc && (*argv)[i + 1] != NULL) i++;
+	av = g_malloc (sizeof(char*) * (argc + 1));
+	av[0] = argv[0];
+
+	for (i = 1; i < argc; i++) {
+		s = argv[i];
+		if (s[0] != '-') continue;
+		if (s[1] == '-') s++;
+		len = strlen (s);
+		if (len > 2 && i + 1 < argc) {
+			if (strncmp ("-display", s, len) == 0) {
+				av[n++] = "--display";
+				av[n++] = argv[++i];
 			}
-			else if (strncmp ("-name", (*argv)[i], len) == 0) {
-				(*argv)[i] = "--name";
-				if (i + 1 < *argc && (*argv)[i + 1] != NULL) i++;
+			else if (strncmp ("-name", s, len) == 0) {
+				av[n++] = "--name";
+				av[n++] = argv[++i];
 			}
 		}
 	}
-
-#if _GTK_2_6
-	/* work around the removal of '-' by gtk_init >= 2.6.0 */
-	for (i = 2; i < *argc; i++) {
-		if (strcmp ((*argv)[i], "-") == 0 &&
-		    (strcmp ((*argv)[i - 1], "-file") == 0 ||
-		     strcmp ((*argv)[i - 1], "--file") == 0)) {
-			(*argv)[i] = "gxmsg:stdin";
-		}
-	}
-#endif
+	av[n] = NULL;
+	ok = gtk_init_check (&n, &av);
+	g_free (av);
+	return ok;
 }
 
 
-
 void
 usage (void)
 {
@@ -825,23 +755,20 @@ usage (void)
 }
 
 
-
 void
 prog_cleanup (void)
 {
-	button_free (gx.button_list);
+	button_free_all (gx.button_list);
 	if (gx.message_text != NULL) {
 		g_free (gx.message_text);
 	}
 	if (gx.timeout_id != 0) {
-		gtk_timeout_remove (gx.timeout_id);
+		g_source_remove (gx.timeout_id);
 	}
-
 	exit (gx.exit_code);
 }
 
 
-
 int
 main (gint argc, gchar *argv[])
 {
@@ -849,48 +776,44 @@ main (gint argc, gchar *argv[])
 	gchar *ch = NULL, *tmpstr;
 	const gchar *fname = NULL;
 	gint opt, arg = 1;
+	gboolean ok;
 	gchar bu_default[] = "okay:0";
 
 	/* The default "okay:0" string is intentionally hard-wired, to avoid
 	 * breaking scripts that make use of xmessage's -print option.
 	 * It must not be changed or gettextize'd. */
 
-
 #ifdef ENABLE_NLS
 	bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
 	bind_textdomain_codeset (PACKAGE, "UTF-8");
 	textdomain (PACKAGE);
 #endif
 
-	option_kludge (&argc, &argv);
-	gtk_init (&argc, &argv);
+	ok = my_gtk_init (argc, argv);
 
 	gx.exit_code       = 1;
 	gx.do_focus        = TRUE;
 	gx.wrap_mode       = GTK_WRAP_NONE;
 	gx.window_position = GTK_WIN_POS_NONE;
 
-
 	while (arg < argc) {
 		opt = my_get_opt (argv[arg], arg + 1 < argc);
 		switch (opt) {
-		case OPT_HELP :
-		case OPT_HELP_Q :
+		case OPT_HELP:
+		case OPT_HELP_Q:
 			gx.exit_code = 0;
 			usage ();
-			break;
-		case OPT_VERSION :
+		case OPT_VERSION:
 			g_print (PACKAGE "-" VERSION "\n");
 			exit (0);
-			break;
-		case OPT_ENTRY :
-		case OPT_ENTRYTEXT :
+		case OPT_ENTRY:
+		case OPT_ENTRYTEXT:
 			if (gx.do_print) {
 				g_printerr (_("%s: can't have both -entry and -print\n"), PACKAGE);
 				prog_cleanup ();
 			}
 			if (gx.timeout) {
-				g_printerr (_("%s: can't have both -entry and -timeout\n"), PACKAGE);
+				/* -entry disables -timeout */
 				gx.timeout = 0;
 			}
 			if (opt == OPT_ENTRY) {
@@ -900,98 +823,97 @@ main (gint argc, gchar *argv[])
 				gx.entry_str = argv[++arg];
 			}
 			break;
-		case OPT_BUTTONS :
-			button_free (gx.button_list);
+		case OPT_BUTTONS:
+			button_free_all (gx.button_list);
 			gx.button_list = button_list_from_str (argv[++arg]);
 			break;
-		case OPT_CENTER :
+		case OPT_CENTER:
 			gx.window_position = GTK_WIN_POS_CENTER;
 			break;
-		case OPT_DEFAULT :
+		case OPT_DEFAULT:
 			gx.default_str = argv[++arg];
 			break;
-		case OPT_FILE :
+		case OPT_FILE:
 			if (gstr != NULL) {
 				g_printerr (_("%s: can't get message from both -file and command line\n"), PACKAGE);
 				prog_cleanup ();
 			}
-			arg++;
-			fname = argv[arg];
+			fname = argv[++arg];
 			break;
-		case OPT_NEARMOUSE :
+		case OPT_NEARMOUSE:
 			/* -center takes priority over -nearmouse */
 			if (gx.window_position != GTK_WIN_POS_CENTER) {
 				gx.window_position = GTK_WIN_POS_MOUSE;
 			}
 			break;
-		case OPT_PRINT :
+		case OPT_PRINT:
 			if (gx.entry_str != NULL) {
 				g_printerr (_("%s: can't have both -entry and -print\n"), PACKAGE);
 				prog_cleanup ();
 			}
 			gx.do_print = TRUE;
 			break;
-		case OPT_TIMEOUT :
+		case OPT_TIMEOUT:
 			gx.timeout = strtol (argv[++arg], &ch, 10);
 			if (*ch) {
 				g_printerr (_("%s: integer -timeout value expected\n"), PACKAGE);
+				/* continue anyway */
 			}
-			if (gx.timeout < 0) {
-				/* xmessage doesn't complain here, just disables timeout */
-				gx.timeout = 0;
-			}
-			if (gx.entry_str != NULL) {
-				g_printerr (_("%s: can't have both -entry and -timeout\n"), PACKAGE);
+			if (gx.timeout < 0 || gx.entry_str != NULL) {
+				/* -entry disables -timeout */
 				gx.timeout = 0;
 			}
 			break;
-		case OPT_TITLE :
+		case OPT_TITLE:
 			gx.title_str = argv[++arg];
 			break;
-		case OPT_GEOMETRY :
+		case OPT_GEOMETRY:
 			gx.geom_str = argv[++arg];
 			break;
-		case OPT_FN :
-		case OPT_FONT :
+		case OPT_FN:
+		case OPT_FONT:
 			gx.font_str = argv[++arg];
 			break;
-		case OPT_RV :
-		case OPT_REVERSE :
-		case OPT_SYNCHRONOUS :
+		case OPT_RV:
+		case OPT_REVERSE:
+		case OPT_SYNCHRONOUS:
 			/* not implemented - ignore */
 			break;
-		case OPT_BG :
+		case OPT_BG:
 			gx.color_bg = argv[++arg];
 			break;
-		case OPT_FG :
+		case OPT_FG:
 			gx.color_fg = argv[++arg];
 			break;
-		case OPT_BD :
-		case OPT_BW :
-		case OPT_XRM :
-		case OPT_SELECTIONTIMEOUT :
-		case OPT_XNLLANGUAGE :
+		case OPT_NAME:
+		case OPT_DISPLAY:
+			/* already handled by my_gtk_init - ignore and skip arg */
+		case OPT_BD:
+		case OPT_BW:
+		case OPT_XRM:
+		case OPT_SELECTIONTIMEOUT:
+		case OPT_XNLLANGUAGE:
 			/* not implemented - ignore and skip arg */
 			arg++;
 			break;
-		case OPT_ICONIC :
+		case OPT_ICONIC:
 			gx.do_iconify = TRUE;
 			break;
-		case OPT_BORDERLESS :
+		case OPT_BORDERLESS:
 			gx.do_borderless = TRUE;
 			break;
-		case OPT_WRAP :
+		case OPT_WRAP:
 			gx.wrap_mode = GTK_WRAP_WORD;
 			break;
-		case OPT_ENCODING :
+		case OPT_ENCODING:
 			gx.encoding = argv[++arg];
 			break;
 		case OPT_FOCUS:
 			gx.do_focus = FALSE;
 			break;
-		case OPT_IS_MISSING_ARG :
+		case OPT_IS_MISSING_ARG:
 			/* in this case, xmessage treats the "option" as normal text */
-		case OPT_IS_UNKNOWN :
+		case OPT_IS_UNKNOWN:
 		default:
 			if (fname != NULL) {
 				g_printerr (_("%s: can't get message from both -file and command line\n"), PACKAGE);
@@ -1009,12 +931,13 @@ main (gint argc, gchar *argv[])
 		arg++;
 	}
 
+	if (!ok) {
+		g_printerr ("%s: unable to initialize GTK\n", PACKAGE);
+		prog_cleanup ();
+	}
+
 	if (fname != NULL) {
-#if _GTK_2_6
-		if (strcmp ("gxmsg:stdin", fname) == 0) {
-#else
 		if (strcmp ("-", fname) == 0) {
-#endif
 			tmpstr = read_stdin ();
 		}
 		else if (!g_file_get_contents (fname, &tmpstr, NULL, NULL)) {
@@ -1022,12 +945,12 @@ main (gint argc, gchar *argv[])
 			prog_cleanup ();
 		}
 		gx.message_text = message_to_utf8 (tmpstr);
-		gx.message_length = strlen (gx.message_text);
+		gx.message_len = strlen (tmpstr);
 		g_free (tmpstr);
 	}
 	else if (gstr != NULL) {
 		gx.message_text = message_to_utf8 (gstr->str);
-		gx.message_length = strlen (gx.message_text);
+		gx.message_len = gstr->len;
 		g_string_free (gstr, TRUE);
 	}
 	else {
@@ -1037,18 +960,14 @@ main (gint argc, gchar *argv[])
 	}
 
 	if (gx.button_list == NULL) {
-		gx.button_list = button_append (NULL,
-		    button_new_from_str (bu_default, 0, strlen (bu_default)));
+		gx.button_list = button_list_from_str (bu_default);
 	}
 
-
 	button_set_default (gx.button_list, gx.default_str);
 
-
 	window_create ();
 	gtk_main ();
 
-
 	prog_cleanup ();
 	return 0;
 }
